/* CorbaApplet.java
 *
 * created: Mon Feb  8 1999
 *
 * This file is part of Artemis
 * 
 * Copyright (C) 1998,1999  Genome Research Limited
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/CorbaApplet.java,v 1.6 1999/03/22 10:59:29 kmr Exp $
 **/

import diana.*;
import diana.components.*;
import diana.sequence.*;

import nsdb.EmblSeq;
import type.NoResult;
import nsdb.EmblPackage.Superceded;

import java.awt.*;
import java.awt.event.*;
import java.net.*;
import java.io.*;
import java.util.*;

/**
 *  This applet will retrieve an embl entry for a corba server and then
 *  display it using a diana.components.EntryEdit object.
 *
 *  @author Kim Rutherford
 *  @version $Id: CorbaApplet.java,v 1.6 1999/03/22 10:59:29 kmr Exp $
 **/

public class CorbaApplet extends java.applet.Applet {
  /**
   *  Initialise this CorbaApplet applet.
   **/
  public void init () {
    Diana.setCodeDirectory (new URLDocument (getCodeBase ()));

    Diana.setIsApplet (true);

    final Font font = Diana.getOptions ().getFont ();

    System.out.println (getCodeBase ());
    System.out.println (Diana.getCodeDirectory ());

    setLayout (new BorderLayout ());

    final Label label = new Label ("Enter an accession number:");
    add (label, "North");

    text_field = new TextField ();
    add (text_field, "Center");

    errors = new Label ("");
    add (errors, "South");

    text_field.addKeyListener (new KeyAdapter () {
      public void keyTyped(final KeyEvent e) {
        if (e.getKeyChar () == '\n') {
          if (text_field.getText () == null ||
              text_field.getText ().length () == 0) {
            return;
          }

          final embl.Entry new_embl_entry = readEntry (text_field.getText ());

          if (new_embl_entry == null) {
            return;
          }

          final Bases embl_bases = new Bases (new_embl_entry.getSequence ());
      
          final EntryGroup entry_group = new EntryGroup (embl_bases);

          new Entry (entry_group.getBases (), new_embl_entry);

          final EntryEdit entry_edit = new EntryEdit (entry_group);

          // this call is done so that the EntryEdit components can be closed
          // later
          Diana.addEntryEdit (entry_edit);

          entry_edit.setSize (700,600);
          entry_edit.validate ();

          errors.setText ("");

          entry_edit.setVisible (true);
        }
      }
    });
  }

  /**
   *  Read and return an embl.Entry object from the server.
   *  @param accession_number The accession number of the entry to read from
   *    the server.  It must be in the same directory as the java code.
   **/
  private embl.Entry readEntry (final String accession_number) {
    final org.omg.CORBA.ORB orb =
      org.omg.CORBA.ORB.init (new String [0], new java.util.Properties ());

    final org.omg.CORBA.Object obj;

    try {
      obj = orb.string_to_object (getIOR ());
    } catch (IOException e) {
      final String message =
        "Error while reading EMBL IOR: " + e.getMessage ();
      errors.setText (message);
      return null;
    }
    
    final nsdb.Embl corba_handle = nsdb.EmblHelper.narrow (obj);

    if (corba_handle == null) {
      final String message = "Failed to get an object from Corba/EMBL";
      errors.setText (message);
      return null;
    }
          
    try {
      final EmblSeq embl_seq = corba_handle.getEmblSeq (accession_number);
      
      return new embl.CorbaEntry (embl_seq);
    } catch (NoResult e) {
      final String message =
        "Database query failed while getting accession number: " +
        accession_number;
      errors.setText (message);
    } catch (Superceded e) {
      //  Superceded is used by getEmblSeq method if accession number
      //  doesn't exist anymore because it was merged or split
      final String message =
        "This accession number has been superceded: " + accession_number;
      errors.setText (message);
    } catch (embl.LocationParseException e) {
      final String message =
        "Unexpected error while accessing " + accession_number + ": " + e;
      errors.setText (message);
    } catch (embl.InvalidKeyException e) {
      final String message =
        "Unexpected error while accessing " + accession_number + ": " + e;
      errors.setText (message);
    }
//      } catch (java.lang.Throwable e) {
//        final String message =
//          "Unexpected error while accessing " + accession_number + ": " + e;
//        errors.setText (message);
//        new MessageDialog (Diana.this, message);
//        return;
//      }

    return null;
  }

  /**
   *  Finds and returns a stringified object reference (IOR).
   *
   *  @return The stringified IOR.
   *  @exception IOException Thrown if an error occurs while reading.
   **/
  private String getIOR ()
      throws IOException {
    try {
      //  get URL of IOR
      URL url = new URL (default_ior_url);
      BufferedReader in =
        new BufferedReader (new InputStreamReader (url.openStream()));
      return in.readLine();
    } catch (MalformedURLException e) {
      throw new Error ("internal error - unexpected MalformedURLException: " +
                       e.getMessage ());
    }
  }

  /**
   *  The default place to look for the IOR for the corba server.
   **/
  static private final String default_ior_url =
    "http://corba.ebi.ac.uk/EMBL/IOR/Embl.IOR";

  /**
   *  A label for status and error messages.
   **/
  private Label errors;

  /**
   *  The user will type the accession number into this field.
   **/
  private TextField text_field;
}
