/* FileProgressInputStream.java
 *
 * created: Thu Feb 25 1999
 *
 * This file is part of Artemis
 * 
 * Copyright (C) 1998,1999  Genome Research Limited
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/FileProgressInputStream.java,v 1.2 1999/03/22 10:59:32 kmr Exp $
 **/

import diana.components.*;

import java.io.*;
import java.awt.*;

/**
 *  This is a FileInputStream which has the readLine () and close () methods
 *  overridden to show the progress in a Label component.
 *
 *  @author Kim Rutherford
 *  @version $Id: FileProgressInputStream.java,v 1.2 1999/03/22 10:59:32 kmr Exp $
 **/

public class FileProgressInputStream extends FileInputStream {
  /**
   *  Creates an input file stream to read from the specified File object and
   *  show the progress in a Frame as it goes.
   *  @param file the file to be opened for reading.
   *  @param progress_label this Label object will be updated to show the
   *    progress of the read with messages like this:
   *    "bytes read so far: 150000"
   *  @exception FileNotFoundException if the file is not found.
   *  @exception SecurityException if a security manager exists, its checkRead
   *    method is called with the pathname of this File argument to see if
   *    the application is allowed read access to the file.
   **/
  public FileProgressInputStream (final File file, final Label progress_label)
      throws FileNotFoundException {
    super (file);
    message = progress_label;
  }

  /**
   * Reads up to <code>b.length</code> bytes of data from this input
   * stream into an array of bytes. This method blocks until some input
   * is available.
   *
   * @param      b   the buffer into which the data is read.
   * @return     the total number of bytes read into the buffer, or
   *             <code>-1</code> if there is no more data because the end of
   *             the file has been reached.
   * @exception  IOException  if an I/O error occurs.
   */
  public int read(byte b[]) throws IOException {
    final int result = super.read (b);

    if (result > 0) {
      byte_count += b.length;
      updateMessage ();
    }

    return result;
  }

  /**
   * Reads up to <code>len</code> bytes of data from this input stream
   * into an array of bytes. This method blocks until some input is
   * available.
   *
   * @param      b     the buffer into which the data is read.
   * @param      off   the start offset of the data.
   * @param      len   the maximum number of bytes read.
   * @return     the total number of bytes read into the buffer, or
   *             <code>-1</code> if there is no more data because the end of
   *             the file has been reached.
   * @exception  IOException  if an I/O error occurs.
   */
  public int read(byte b[], int off, int len) throws IOException {
    final int result = super.read (b, off, len);

    if (result > 0) {
      byte_count += b.length;
      updateMessage ();
    }

    return result;
  }

  /**
   *  Close this file input stream, close the progress indicator component
   *  and release any system resources associated with the stream.
   *
   *  @exception  IOException  if an I/O error occurs.
   **/
  public void close ()
      throws IOException {
    super.close ();

    message.setText ("");
  }

  /**
   *  Update the progress indicator to show how many byte have have read so
   *  far.
   **/
  private void updateMessage () {
    if (byte_count > 100000) {
      if (last_count_message + 50000 < byte_count) {
        final int rounded_byte_count = (byte_count / 50000) * 50000;

        message.setText ("bytes read so far: " + rounded_byte_count);
        last_count_message = byte_count;
      }
    }
  }

  /**
   *  This is the component used to show the progress.
   **/
  private final Label message;

  /**
   *  The number of bytes that have been read so far.
   **/
  private int byte_count = 0;

  /**
   *  This this byte count that was shown last time in the message.
   **/
  private int last_count_message = 0;
}
