/* ExternalSearchProgram.java
 *
 * created: Tue Jan 26 1999
 *
 * This file is part of Artemis
 * 
 * Copyright (C) 1998,1999  Genome Research Limited
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/diana/ExternalSearchProgram.java,v 1.10 1999/03/22 11:00:01 kmr Exp $
 **/

package diana;

import FileDocument;
import Document;
import Diana;
import StringVector;
import diana.components.*;

import java.io.*;
import java.text.*;

/**
 *  This class is an ExternalProgram that can run searches.
 *
 *  @author Kim Rutherford
 *  @version $Id: ExternalSearchProgram.java,v 1.10 1999/03/22 11:00:01 kmr Exp $
 **/

public class ExternalSearchProgram extends ExternalProgram {
  /**
   *  Create a new ExternalSearchProgram object.
   *  @param name The name of the program.
   *  @param search_database The default database to run the program on.
   **/
  public ExternalSearchProgram (final String name,
                                final String search_database) {
    super (name);

    this.search_database = search_database;
  }

  /**
   *  Run this program with the options that were set with setOptions ().
   *  @param features The program will be run for each of these features.
   *  @exception IOException Thrown if an IO error occur while trying to run
   *    the program (eg the program could not be found.
   *  @exception ExternalProgramException Thrown if there is a non-IO error
   *    while attempting to run the program.
   **/
  public void run (final FeatureVector features)
      throws IOException, ExternalProgramException {

    try {
      if (features.size () == 0) {
        return;
      }

      final FileDocument code_directory = getCodeDirectory ();

      if (code_directory == null) {
        throw new ExternalProgramException ("internal error - " +
                                            "there is no code directory");
      }

      makeDirectory (new File (getName ()));

      final StringVector sequence_file_names = new StringVector ();

      final NumberFormat number_format = NumberFormat.getNumberInstance ();

      number_format.setMaximumIntegerDigits (5);
      number_format.setMinimumIntegerDigits (5);
      number_format.setGroupingUsed (false);

      // this number is incremented each time a sequence file is written, and
      // is stored a file in the current directory
      int file_number = getFileNumber ();

      // write the sequences out
      for (int i = 0 ; i < features.size () ; ++i) {
        final Feature feature = features.elementAt (i);
        final String entry_name;

        {
          final String test_name = feature.getEntry ().getName ();
          if (test_name == null) {
            entry_name = "no_name";
          } else {
            entry_name = test_name;
          }
        }

        final String new_file_name =
          getName () + File.separatorChar + entry_name + ".seq." +
          number_format.format (file_number);

        sequence_file_names.add (new_file_name);

        final Writer writer = new FileWriter (new File (new_file_name));

        feature.writePIROfFeature (writer);

        writer.close ();

        final embl.Qualifier new_qualifier =
          new embl.Qualifier (getName () + "_file",
                              new_file_name + ".out");
        feature.setQualifier (new_qualifier);

        ++file_number;
      }

      setFileNumber (file_number);

//      makeDirectory (getName () + File.pathSeparator +
//                     getSearchDatabase ())

      // write the sequence file names to a file and then run the program with
      // that file as an argument
      final String file_of_filenames_name = "file_of_filenames." + getName ();

      final Writer filenames_writer = new FileWriter (file_of_filenames_name);

      final PrintWriter filenames_printwriter =
        new PrintWriter (filenames_writer);

      for (int i = 0 ; i < sequence_file_names.size () ; ++i) {
        filenames_printwriter.println (sequence_file_names.elementAt (i));
      }

      filenames_printwriter.close ();
      filenames_writer.close ();

      final String program_file_name =
        code_directory.getFile ().getAbsolutePath () +
        File.separatorChar + "run_" + getName ();

      final String [] arguments = {
        program_file_name,
        "-f",
        file_of_filenames_name,
        getSearchDatabase ()
      };

      final String [] envp = {
        "DIANA_ENVIRONMENT_FILE=" + System.getProperty ("diana.environment")
      };

      final Process process =
        Runtime.getRuntime ().exec (arguments, envp);

      new MessageFrame ("started " + getName () + " on " + features.size () +
                        (features.size () == 1 ? " feature" : " features"));
    } catch (SecurityException e) {
      // re-throw as an ExternalProgramException
      throw new ExternalProgramException ("SecurityException while running " +
                                          getName () + ": " + e.getMessage ());
    }
  }

  /**
   *  This method will popup a form with the options for this ExternalProgram
   *  object.  For a ExternalSearchProgram there is only one option to set -
   *  the database.
   **/
  public void setOptions () {
    final TextRequester requester =
      new TextRequester ("Choose a database for " + getName () + ":", 18,
                         getSearchDatabase ());

    requester.addTextRequesterListener (new TextRequesterListener () {
      public void actionPerformed (final TextRequesterEvent event) {
        final String requester_text = event.getRequesterText ().trim ();
        if (requester_text.length () > 0) {
          setSearchDatabase (requester_text);
        }
      }
    });

    requester.show ();
  }

  /**
   *  Return the FileDocument of the directory where the program scripts
   *  should be or null if the directory is unset.
   **/
  private FileDocument getCodeDirectory () {
    final Document code_directory = Diana.getCodeDirectory ();

    if (code_directory instanceof FileDocument) {
      return (FileDocument) code_directory;
    } else {
      return null;
    }
  }

  /**
   *  Return the name of the database that will be used when the program is
   *  run.
   **/
  public String getSearchDatabase () {
    return search_database;
  }

  /**
   *  Set the name of the database that will be used when the program is run.
   **/
  public void setSearchDatabase (final String search_database) {
    this.search_database = search_database;
  }

  /**
   *  The name of the file that stores the counter used when creating the
   *  protein files to run the search on.  The file contains a comment line
   *  and then a line with a single integer on it.  This file is updated by
   *  setFileNumber () and read with getFileNumber ().
   **/
  final static String file_counter_filename = "file_number_counter";

  /**
   *  Read the number to be used with creating the next protein output file.
   *  The number is read from file_counter_filename.
   **/
  private int getFileNumber ()
      throws IOException {
    try {
      final FileReader file_reader =
        new FileReader (getName () + File.separatorChar +
                        file_counter_filename);

      final BufferedReader reader = new BufferedReader (file_reader);

      // the first line is a comment
      reader.readLine ();

      final String number_line = reader.readLine ();

      try {
        return Integer.parseInt (number_line);
      } catch (NumberFormatException e) {
        new MessageFrame ("the file_number_counter file is corrupted - " +
                        "delete it and try again");
        return 1;
      } finally {
        file_reader.close ();
        reader.close ();
      }
    } catch (FileNotFoundException e) {
      // create a new file_number_counter file
      setFileNumber (1);
      return 1;
    }
  }

  /**
   *  Write the given number to the file given by new_file_name.
   **/
  private void setFileNumber (final int new_file_number) {
    try {
      final FileWriter file_writer =
        new FileWriter (getName () + File.separatorChar +
                        file_counter_filename);

      final PrintWriter print_writer = new PrintWriter (file_writer);

      print_writer.println ("# the file is machine generated - do not edit");

      print_writer.println (new_file_number);

      print_writer.close ();

      file_writer.close ();
    } catch (IOException e) {
      new MessageFrame ("unexpected exception while writing " +
                        file_counter_filename + ": " + e.getMessage ());
    }
  }

  /**
   *  Create a directory path with the given directory_name (if it does not
   *  exist).
   **/
  private void makeDirectory (final File directory_name)
      throws IOException {
    if (directory_name.exists ()) {
      if (directory_name.isDirectory ()) {
        if (directory_name.canWrite ()) {
          // all is ok
          return;
        } else {
          throw new IOException ("Cannot write to: " +
                                 directory_name.getAbsolutePath ());
        }
      } else {
        throw new IOException (directory_name.getAbsolutePath () +
                               " is not a directory");
      }
    } else {
      directory_name.mkdirs ();
    }
  }

  /**
   *  The default database to use for searches with this program.
   **/
  private String search_database;
}


