/* Options.java
 *
 * created: Thu Dec 10 1998
 *
 * This file is part of Artemis
 * 
 * Copyright (C) 1998,1999  Genome Research Limited
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/diana/Options.java,v 1.20 1999/03/30 15:03:24 kmr Exp $
 **/

package diana;

import Document;
import StringVector;
import embl.Key;

import java.awt.*;
import java.io.*;
import java.util.*;

/**
 *  An object of this class is used to read, write and store all the
 *  configurable options in Diana.  The method Properties.load () is used to
 *  read in new options.
 *
 *  @author Kim Rutherford
 *  @version $Id: Options.java,v 1.20 1999/03/30 15:03:24 kmr Exp $
 **/

public class Options extends Properties {
  /**
   *  Create a new Options object with default settings for the options.
   **/
  public Options () {
    put ("font_size", "14");
    setDefaultFeatureColours ();
    setDefaultColourMap ();

    resetCachedValues ();
  }

  /**
   *  Call load () from the super class and then reset the cached values for
   *  Font and Color objects.
   **/
  public void load (InputStream in_stream)
      throws IOException {
    super.load (in_stream);

    resetCachedValues ();
  }

  /**
   *  Return a StringVector containing the values of the given option or null
   *  if there is no such option.  The values are separated by spaces in the
   *  properties file.
   *
   *  @see #getPropertyValues ()
   **/
  public StringVector getOptionValues (final String option) {
    return getPropertyValues (this, option);
  }

  /**
   *  Return a StringVector containing the values of the given option in the
   *  given Properties object or null if there is no such option.  The values
   *  are separated by spaces in the properties file.  For example, asking for
   *  the property foo in a Properties object where foo has the value: "thing1
   *  thing2 thing3", will return a vector containing three String objects:
   *  "thing1", "thing2", "thing3"
   **/
  public static StringVector getPropertyValues (final Properties properties,
                                                final String option) {
    final String option_value = properties.getProperty (option);

    if (option_value == null) {
      return null;
    }

    return StringVector.getStrings (option_value);
  }

  /**
   *  Return a QualifierInfoVector object containing one object for each
   *  qualifier given in extra_qualifiers option.
   *  @exception QualifierInfoException Thrown if any of the type string do not
   *    contain a valid type string.
   **/
  public embl.QualifierInfoVector getExtraQualifiers ()
      throws embl.QualifierInfoException {
    final embl.QualifierInfoVector return_vector =
      new embl.QualifierInfoVector ();

    final StringVector extra_qualifiers_strings =
      getOptionValues ("extra_qualifiers");

    for (int i = 0 ; i < extra_qualifiers_strings.size () / 2 ; ++i) {
      final String name = extra_qualifiers_strings.elementAt (i * 2);
      final String type = extra_qualifiers_strings.elementAt (i * 2 + 1);

      return_vector.add (new embl.QualifierInfo (name, type));
    }

    return return_vector;
  }

  /**
   *  Return the path of the default sequence file or null if there is no
   *  default.  This will usually have been set in the diana.ini file,
   *  eg. SEQUENCE_FILE='c5H10.seq'
   **/
  public String getDefaultSequenceFileName () {
    final String property_string = getProperty ("SEQUENCE_FILE");

    if (property_string == null) {
      return null;
    } else {
      // trim the quotes from the name
      if (property_string.startsWith ("'") &&
          property_string.endsWith ("'") &&
          property_string.length () > 2) {
        return property_string.substring (1, property_string.length () - 1);
      } else {
        return property_string;
      }
    }
  }

  /**
   *  Return the path of the default feature file or null if there is no
   *  default.  This will usually have been set in the diana.ini file,
   *  eg. FEATURE_FILE='c5H10_embl.tab'
   **/
  public String getDefaultFeatureFileName () {
    final String property_string = getProperty ("FEATURE_FILE");

    if (property_string == null) {
      return null;
    } else {
      // trim the quotes from the name
      if (property_string.startsWith ("'") &&
          property_string.endsWith ("'") &&
          property_string.length () > 2) {
        return property_string.substring (1, property_string.length () - 1);
      } else {
        return property_string;
      }
    }
  }

  /**
   *  Return the minimum open reading frame size used when automatically
   *  creating ORFs with the "Mark Open Reading Frames" function.
   **/
  public int getMinimumORFSize () {
    final Integer minimum_orf_size = getIntegerProperty ("minimum_orf_size");

    if (minimum_orf_size == null) {
      // default value
      return 100;
    } else {
      return minimum_orf_size.intValue ();
    }
  }

  /**
   *  Get the default colour for the given feature key, as specified in the
   *  options file.  If no colour is specified in the file then null is
   *  returned.
   *  @param key The feature key to get the colour of.
   **/
  public Color getDefaultFeatureColour (Key key) {
    final Integer colour_integer = getIntegerProperty ("colour_of_" + key);

    if (colour_integer == null) {
      return null;
    } else {
      return getColorFromColourNumber (colour_integer.intValue ());
    }
  }

  /**
   *  Given a colour number (perhaps from a /colour qualifier) return an
   *  appropriate Color object.
   **/
  public Color getColorFromColourNumber (int colour_number) {
    // first look up colour_map for speed.  if that fails we then try to look
    // for an appropriate property and turn into a Color object.
    if (colour_number < 0) {
      return null;
    }

    if (colour_number >= colour_map.size () ||
        colour_map.elementAt (colour_number) == null) {
      // there is no colour for this colour_number
      return null;
    } else {
      return (Color) colour_map.elementAt (colour_number);
    }
  }

  /**
   *  Given a String containing three space separated integers (0-255), return
   *  a Color object.  The three integers represent red, green and blue
   *  respectively.  If a Color can't be parsed null is returned.
   **/
  private Color parseColour (String colour_string) {
    try {
      // first get three integers from the String

      // trim any whitespace from the ends
      final String real_colour_string = colour_string.trim ();

      final int first_space = real_colour_string.indexOf (' ');

      if (first_space == -1) {
        return null;
      }

      final int first_int =
        Integer.valueOf (real_colour_string.substring (0, first_space)).intValue ();

      final String rest_of_string =
        real_colour_string.substring (first_space + 1).trim ();

      final int second_space = rest_of_string.lastIndexOf (' ');

      final int second_int =
        Integer.valueOf (rest_of_string.substring (0, second_space)).intValue ();
      final int third_int =
        Integer.valueOf (rest_of_string.substring (second_space + 1)).intValue ();

      return new Color (first_int, second_int, third_int);

    } catch (NumberFormatException e) {
      return null;
    }
  }

  /**
   *  Get the default Color for a colour
   **/
  public void setDefaultColourMap () {
    put ("colour_0", "255 255 255"); // white
    put ("colour_1", "100 100 100"); // dark grey
    put ("colour_2", "255 0 0");     // red
    put ("colour_3", "0 255 0");     // green
    put ("colour_4", "0 0 255");     // blue
    put ("colour_5", "0 255 255");   // cyan
    put ("colour_6", "255 0 255");    // magenta
    put ("colour_7", "255 255 0");    // yellow
    put ("colour_8", "152 251 152");  // pale green
    put ("colour_9", "135 206 250");  // light sky blue
    put ("colour_10", "255 165 0");    // orange
  }

  /**
   *  Return a vector containing the ExternalProgram objects of all the
   *  external programs that we can use.
   **/
  public ExternalProgramVector getExternalPrograms () {
    if (external_programs == null) {
      external_programs = new ExternalProgramVector ();

      external_programs.add (new ExternalSearchProgram ("fasta", "%SNTU"));
      external_programs.add (new ExternalSearchProgram ("blast", "swir"));
    }

    return external_programs;
  }

  /**
   *  Return a String containing the possible start codons.  This is stored in
   *  the start_codons option in the options file.
   **/
  public String getStartCodonSymbols () {
    return getProperty ("start_codons").toLowerCase ();
  }

  /**
   *  Return the default font that should be used for all windows.
   **/
  public Font getFont () {
    return font;
  }

  /**
   *  Set the default colour number for each a feature key.
   **/
  private void setDefaultFeatureColours () {
    final Object [] key_colour_map = {
      "CDS", "5",
      "cds?", "7",
      "BLASTCDS", "2",
      "BLASTN_HIT", "6",
      "source", "0",
      "prim_tran", "0",
      "stem_loop", "2",
      "misc_feature", "3",
      "delta", "3",
      "repeat_region", "9",
      "repeat_unit", "9",
      "terminator", "3",
      "promoter", "3",
      "intron", "1",
      "exon", "7",
      "mRNA", "1",
      "tRNA", "8",
      "TATA", "3",
      "bldA", "2"
    };

    for (int i = 0 ; i < key_colour_map.length / 2 ; ++i) {
      put ("colour_of_" + key_colour_map[i*2], key_colour_map[i*2+1]);
    }
  }

  /**
   *  Clear all cached values (such as the font) and then recalculate.
   **/
  private void resetCachedValues () {
    final int font_size;

    try {

      final Integer font_size_integer = getIntegerProperty ("font_size");
      font_size = font_size_integer.intValue ();

    } catch (NumberFormatException e) {
      new diana.components.MessageFrame ("error in options file - " +
                                         "font_size should be an integer");
      // a default value
      font_size = 14;
      put ("font_size", String.valueOf (font_size));
    }

    font = new Font ("Monospaced", Font.PLAIN, font_size);

    colour_map = new Vector (25);

    int colour_number = 0;

    while (true) {

      final String colour_value_string =
        getProperty ("colour_" + colour_number);

      if (colour_value_string == null) {
        // we know nothing about this colour number so we assume that there
        // aren't any more
        break;
      } else {
        final Color new_colour = parseColour (colour_value_string);

        if (new_colour == null) {
          // we couldn't parse the colour
          new diana.components.MessageFrame ("could not understand this " +
                                             "number: " +
                                             colour_value_string);
        } else {
          if (colour_number >= colour_map.size ()) {
            colour_map.setSize (colour_number + 50);
          }
          colour_map.setElementAt (new_colour, colour_number);
        }
      }

      ++colour_number;
    }
  }

  /**
   *  Return the value of a property/option as an Integer.  If the property
   *  does not exist or cannot be parsed as an Integer, the method returns
   *  null.
   **/
  public Integer getIntegerProperty (String property_name) {
    final String property_string = getProperty (property_name);

    if (property_string == null) {
      return null;
    } else {
      return Integer.valueOf (property_string);
    }
  }

  /**
   *  The default font that should be used for all windows.
   **/
  private Font font = null;

  /**
   *  A map of colour numbers to Color object.  This is initialised by
   *  setDefaultColourMap ().
   **/
  private Vector colour_map = null;

  /**
   *  A vector containing the ExternalProgram objects that we can run.
   **/
  private ExternalProgramVector external_programs = null;
}

