/* EntryEdit.java
 *
 * created: Fri Oct  9 1998
 *
 * This file is part of Artemis
 *
 * Copyright (C) 1998,1999  Genome Research Limited
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/diana/components/EntryEdit.java,v 1.78 1999/05/04 10:00:31 kmr Exp $
 */

package diana.components;

import Diana;
import OutOfRangeException;
import ReadOnlyException;

import diana.*;
import diana.plot.*;
import diana.sequence.*;

import java.awt.*;
import java.awt.event.*;
import java.util.Vector;
import java.util.EventObject;
import java.io.IOException;
import java.io.File;
import java.io.FileReader;

/**
 *  Each object of this class is used to edit an EntryGroup object.
 *
 *  @author Kim Rutherford
 *  @version $Id: EntryEdit.java,v 1.78 1999/05/04 10:00:31 kmr Exp $
 *
 */

public class EntryEdit extends Frame
    implements SelectionOwner, GotoEventSource {
  /**
   *  Create a new EntryEdit object and Frame.
   *  @param entry_group The EntryGroup object that this component is editing.
   */
  public EntryEdit (EntryGroup entry_group) {
    super ("P.O.W.M.A.P. Entry Edit");

    this.entry_group = entry_group;

    getEntryGroup ().addFeatureChangeListener (selection);
    getEntryGroup ().addEntryChangeListener (selection);

    if (getEntryGroup ().getDefaultEntry () != null) {
      final String name = getEntryGroup ().getDefaultEntry ().getName ();

      if (name != null) {
        setTitle ("P.O.W.M.A.P. Entry Edit: " + name);
      }
    }

    final int font_height;
    final int font_base_line;

    final Font default_font = getDefaultFont ();

    setFont (default_font);

    if (default_font != null) {
      FontMetrics fm = getFontMetrics (default_font);
      //      System.out.println ("fm:" + fm.toString ());
      font_height = fm.getHeight ();
      font_base_line = fm.getMaxAscent ();
    }
    else {
      font_height = -1;
      font_base_line = -1;
    }

//    System.out.println ("EntryEdit (): " + entry_group);

    addWindowListener (new WindowAdapter () {
      public void windowClosing (WindowEvent event) {
        entryEditFinished ();
      }
    });

    GridBagLayout gridbag = new GridBagLayout();
    setLayout (gridbag);

    menu_bar.setFont (default_font);

    GridBagConstraints c = new GridBagConstraints();

    c.gridwidth = GridBagConstraints.REMAINDER;
    c.fill = GridBagConstraints.HORIZONTAL;
    c.anchor = GridBagConstraints.NORTH;
    c.gridheight = 1;
    c.weightx = 1;
    c.weighty = 0;

    // Create and add the stats_label.
    selection_info = new SelectionInfoDisplay (this);
    gridbag.setConstraints (selection_info, c);
    add (selection_info);

    group_display = new EntryGroupDisplay (this);

    gridbag.setConstraints (group_display, c);
    add (group_display);

    feature_display = new FeatureDisplay (this);

    base_plot_group = new BasePlotGroup (this);
    gridbag.setConstraints (base_plot_group, c);
    add (base_plot_group);
    feature_display.addDisplayAdjustmentListener (base_plot_group);

    gridbag.setConstraints (feature_display, c);
    add (feature_display);

    base_display = new FeatureDisplay (this);
    base_display.setShowLabels (false);
    base_display.setScaleFactor (0);
    gridbag.setConstraints (base_display, c);
    add (base_display);

    final String list_option = Diana.getOptions ().getProperty ("show_list");

    feature_list = new FeatureList (this);
    //    feature_list.setBackground (Color.gray);
    feature_list.setFont (default_font);
    c.fill = GridBagConstraints.BOTH;
    c.weighty = 1;
    gridbag.setConstraints (feature_list, c);

    if (list_option == null || ! list_option.equals ("no")) {
      feature_list.setVisible (true);
    } else {
      feature_list.setVisible (false);
    }

    add (feature_list);

    makeMenus ();

    packme ();

    setSize (900,750);

    final Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
    setLocation (new Point ((screen.width - getSize ().width) / 2,
                            (screen.height - getSize ().height) / 2));
  }

  /**
   *  Redraw this component.  This method is public so that other classes can
   *  force an update if for example the options files is re-read.
   **/
  public void redisplay () {
    feature_display.redisplay ();
    base_display.redisplay ();
  }

  /**
   *  Add an Entry to this object.  This causes the features from the entry to
   *  be displayed in the list and adds this EntryEdit object as a listener
   *  for EntryChangeEvent objects from the Entry.
   *  @see #getEntryGroup ()
   *  @param entry The Entry object to add.
   **/
  public void add (Entry entry) {
    entry_group.add (entry);
    Thread.yield ();
  }

  /**
   *  This method sends an GotoEvent to all the GotoEvent listeners that will
   *  make the given base visible.
   *  @param base_marker The base to make visible.
   **/
  public void gotoBase (final Marker base_marker) {
    final GotoEvent new_event = new GotoEvent (this, base_marker);

    sendGotoEvent (new_event);
  }

  /**
   *  This method sends an GotoEvent to all the GotoEvent listeners that will
   *  make the start of the first selected Feature or FeatureSegment visible.
   **/
  public void makeSelectionVisible () {
    final Marker first_base = getSelection ().getFirstBaseOfSelection ();

    final GotoEvent new_event = new GotoEvent (this, first_base);

    sendGotoEvent (new_event);
  }

  /**
   *  Send the given event to all the GotoListeners.
   **/
  public void sendGotoEvent (final GotoEvent goto_event) {
    // don't send if there is nowhere to go to
    if (goto_event.getMarker () != null) {
      fireAction (goto_listener_list, goto_event);
    }
  }

  /**
   *  This method sends an GotoEvent to all the GotoEvent listeners that will
   *  make the given base visible.  This just calls makeBaseVisible ().
   *  @param destination_base The base to make visible.
   *  @return The reference of a MarkerRange object for the given base or null
   *    if the call fails.  It will fail only if the destination_base is less
   *    than 1 or greater than the length of the sequence.
   **/
  public MarkerRange gotoBase (final int destination_base) {
    try {
      final Strand forward_strand =
        getEntryGroup ().getBases ().getForwardStrand ();
      final Marker destination_marker =
        forward_strand.makeMarker (destination_base);
      gotoBase (destination_marker);
      return new MarkerRange (destination_marker.getStrand (),
                              destination_base, destination_base);
    } catch (OutOfRangeException e) {
      return null;
    }
  }

  /**
   *  This method sends a GotoEvent to all the GotoEvent listeners that will
   *  make the first base of the sequence visible.
   **/
  public void gotoFirstBase () {
    gotoBase (1);
  }

  /**
   *  This method sends a GotoEvent to all the GotoEvent listeners that will
   *  make the last base of the sequence visible.
   **/
  public void gotoLastBase () {
    gotoBase (getEntryGroup ().getSequenceLength ());
  }

  /**
   *  This method finds the given pattern and then sends an GotoEvent to all
   *  the GotoEvent listeners that will make the first base of the match
   *  visible.
   *  @param pattern This is the pattern to search for.
   *  @param start_at_first_base If true or if there is nothing in the
   *    Selection then the search will start at base 1, otherwise the search
   *    will start at the Selection.
   *  @return The range that matches the pattern or null if there is no match.
   **/
  public MarkerRange gotoBasePattern (final BasePattern pattern,
                                      boolean start_at_first_base) {
    // if start_at_first_base is false we want to start the search at the
    // selection
    final Marker selection_base;

    final Marker first_selection_base =
      getSelection ().getFirstBaseOfSelection ();

    final Marker last_selection_base =
      getSelection ().getLastBaseOfSelection ();

    // these checks are currently unnecessary, because if first_selection_base
    // is null then last_selection_base will be too.  This is future proofing
    if (first_selection_base == null || last_selection_base == null) {
      if (first_selection_base == null) {
        selection_base = last_selection_base;
      } else {
        selection_base = first_selection_base;
      }
    } else {
      selection_base =
        Marker.getRawLowest (first_selection_base, last_selection_base);
    }

    final Marker start_position;

    if (start_at_first_base || selection_base == null) {
      // null means start the search from the beginning
      start_position = null;
    } else {
      start_position = selection_base;
    }

    final MarkerRange match_range =
      pattern.findMatch (getEntryGroup ().getBases (),
                         start_position,
                         getEntryGroup ().getSequenceLength ());

    if (match_range == null) {
      return null;
    } else {
      gotoBase (match_range.getStart ());
      return match_range;
    }
  }

  /**
   *  Adds the specified event listener to receive Goto events from this
   *  object.
   *  @param l the GotoEvent listener.
   **/
  public void addGotoListener (final GotoListener l) {
    goto_listener_list.addElement (l);
  }

  /**
   *  Removes the specified event listener so that it no longer receives Goto
   *  events from this object.
   *  @param l the GotoEvent listener.
   **/
  public void removeGotoListener (final GotoListener l) {
    goto_listener_list.removeElement (l);
  }

  /**
   *  remove an Entry from this object.  This causes the features from the
   *  entry to removed from the display and removes this EntryEdit object as a
   *  listener for EntryChangeEvent objects from the Entry.
   *  @param entry The Entry object to remove.
   **/
  public void remove (Entry entry) {
    entry_group.remove (entry);
    Thread.yield ();
  }

  /**
   *  Return the EntryGroup object that was passed to the constructor.  The
   *  EntryGroup can be changed by calling the methods EntryEdit.add () and
   *  EntryEdit.remove ().  An Object that needs to monitor changes to the
   *  EntryGroup should add itself as a listener to the EntryGroup.
   **/
  public EntryGroup getEntryGroup () {
    return entry_group;
  }

  /**
   *  Returns a Selection object containing the selected features/exons.
   **/
  public Selection getSelection () {
    return selection;
  }

  /**
   *  Read a new entry into the entry group that this component is displaying.
   *  This method uses EntryFileDialog to open a requester.
   **/
  public void readEntry () {
    final EntryFileDialog dialog =
      new EntryFileDialog (this);

    // get the entry and add it to entry_group

    final embl.Entry new_embl_entry = dialog.getEntry ();

    if (new_embl_entry == null) {
      // the read failed
      return;
    }

    final Bases embl_bases = new Bases (new_embl_entry.getSequence ());

    final Entry new_entry =
      new Entry (entry_group.getBases (), new_embl_entry);

    getEntryGroup ().add (new_entry);
  }

  /* private members: */

  /**
   *  This method arranges for the EntryEdit Frame to go away.  This EntryEdit
   *  object was created by the main program, so the main program must be the
   *  one to delete us.
   **/
  private void entryEditFinished () {
    Diana.entryEditFinished (this);
  }

  /**
   *  Send an event to those object listening for it.
   *  @param listeners A Vector of the objects that the event should be sent
   *    to.
   *  @param event The event to send
   **/
  private void fireAction (final Vector listeners, final EventObject event) {
    final Vector targets;
    // copied from a book - synchronising the whole method might cause a
    // deadlock
    synchronized (this) {
      targets = (Vector) listeners.clone ();
    }

    for (int i = 0 ; i < targets.size () ; ++i) {
      GotoListener target = (GotoListener) targets.elementAt (i);

      if (event instanceof GotoEvent) {
        final GotoListener goto_listener = (GotoListener) target;
        goto_listener.performGoto ((GotoEvent) event);
      } else {
        throw new Error ("EntryEdit.fireAction () - unknown event");
      }
    }
  }

  /**
   *  Ask the user for a new default Entry then set it.
   **/
  private void setDefaultEntry () {
    final EntryChooser entry_chooser =
      new EntryChooser (entry_group, "choose a default entry:");

    final EntryChooserListener new_listener =
      new EntryChooserListener () {
        public void entryChoosen (EntryChooserEvent event) {
          entry_group.setDefaultEntry (event.getChoosenEntry ());
          entry_chooser.dispose ();
        }
      };

    entry_chooser.addEntryChooserListener (new_listener);

    entry_chooser.setVisible (true);
  }

  /**
   *  Write an Entry from the EntryGroup to the file it came from.  An
   *  EntryEdit object can potentially display several Entry objects at once,
   *  so we use an EntryChooser component to choose one to save.
   **/
  private void saveAnEntry () {
    final EntryChooser entry_chooser =
      new EntryChooser (entry_group, "choose an entry to save:");

    final EntryChooserListener new_listener =
      new EntryChooserListener () {
        public void entryChoosen (EntryChooserEvent event) {
          final Entry choosen_entry = event.getChoosenEntry ();
          saveEntry (choosen_entry, true, false, true);
          entry_chooser.dispose ();
        }
      };

    entry_chooser.addEntryChooserListener (new_listener);

    entry_chooser.setVisible (true);
  }

  /**
   *  Write an Entry from the EntryGroup - querying the user for the filename.
   *  An EntryEdit object can potentially display several Entry objects at
   *  once, so we use an EntryChooser component to choose one to save.
   **/
  private void saveAnEntryAs () {
    final EntryChooser entry_chooser =
      new EntryChooser (entry_group, "choose an entry to save:");

    final EntryChooserListener new_listener =
      new EntryChooserListener () {
        public void entryChoosen (EntryChooserEvent event) {
          final Entry choosen_entry = event.getChoosenEntry ();
          saveEntry (choosen_entry, true, true, true);
          entry_chooser.dispose ();
        }
      };

    entry_chooser.addEntryChooserListener (new_listener);

    entry_chooser.setVisible (true);
  }

  /**
   *  Write an Entry from the EntryGroup to the file it came from, removing
   *  non-standard qualifiers etc., in the process.  An EntryEdit object can
   *  potentially display several Entry objects at once, so we use an
   *  EntryChooser component to choose one to save.
   **/
  private void saveAnEntryAsEMBL () {
    final EntryChooser entry_chooser =
      new EntryChooser (entry_group, "choose an entry to save:");

    final EntryChooserListener new_listener =
      new EntryChooserListener () {
        public void entryChoosen (EntryChooserEvent event) {
          final Entry choosen_entry = event.getChoosenEntry ();
          saveEntry (choosen_entry, false, true, false);
          entry_chooser.dispose ();
        }
      };

    entry_chooser.addEntryChooserListener (new_listener);

    entry_chooser.setVisible (true);
  }

  /**
   *  Write the default Entry in the EntryGroup to the file it came from.
   **/
  private void saveDefaultEntry () {
    if (getEntryGroup ().getDefaultEntry () == null) {
      new MessageDialog (EntryEdit.this, "There is no default entry");
    } else {
      saveEntry (entry_group.getDefaultEntry (), true, false, true);
    }
  }

  /**
   *  Create a new TextRequester component and set the name of the default
   *  Entry to whatever the user types.
   **/
  private void setDefaultEntryName () {
    if (getEntryGroup ().getDefaultEntry () == null) {
      new MessageDialog (EntryEdit.this, "There is no default entry");
      return;
    }

    final TextRequester text_requester =
      new TextRequester ("New name for the default entry?", 18, "");

    text_requester.addTextRequesterListener (new TextRequesterListener () {
      public void actionPerformed (final TextRequesterEvent event) {
        final String requester_text = event.getRequesterText ().trim ();
        if (requester_text.length () == 0) {
          new MessageDialog (EntryEdit.this, "name too short");
        } else {
          if (entry_group.getDefaultEntry ().setName (requester_text)) {
            // it worked
          } else {
            new MessageDialog (EntryEdit.this,
                               "could not set the name of the default entry");
          }
        }
      }
    });

    text_requester.show ();
  }

  /**
   *  Save the given entry, prompting for a file name if necessary.
   *  @param include_diana_extensions If true then the any diana additions to
   *    the embl file format will be included in the output, otherwise they
   *    will be removed.  Also possible problems that would cause an entry to
   *    bounce from the EMBL submission process will be flagged if this is
   *    true.
   *  @param ask_for_name If true then always prompt for a new filename,
   *    otherwise prompt only when the entry name is not set.
   *  @param keep_new_name If ask_for_name is true a file will be written with
   *    the new name the user selects - if keep_new_name is true as well, then
   *    the entry will have it's name set to the new name, otherwise it will
   *    be used for this save and then discarded.
   **/
  private void saveEntry (final Entry entry,
                          final boolean include_diana_extensions,
                          final boolean ask_for_name,
                          final boolean keep_new_name) {

    final FeatureVector duplicate_features;

    if (include_diana_extensions) {
      duplicate_features = null;
    } else {
      duplicate_features = entry.checkForEMBLDuplicates ();
    }

    if (duplicate_features != null) {
      if (duplicate_features.size () > 0) {
        final YesNoDialog yes_no_dialog =
          new YesNoDialog (this,
                           "warning: there are features with identical " +
                           "keys and locations - continue?");

        if (yes_no_dialog.getResult ()) {
          // yes - continue
        } else {
          // no - select the duplicates and return;
          final MessageDialog message =
            new MessageDialog (this,
                               "the features with identical " +
                               "keys and locations have been selected",
                               false);

          getSelection ().set (duplicate_features);
          return;
        }
      }
    }

    final FeatureVector invalid_features;

    if (include_diana_extensions) {
      invalid_features = null;
    } else {
      invalid_features = entry.checkForStopsAndStarts ();
    }

    try {
      if (ask_for_name || entry.getName () == null) {
        // prompt for a name
        final FileDialog dialog =
          new FileDialog (this,
                          "Select an EMBL file name...",
                          FileDialog.SAVE);

        dialog.setVisible (true);

        if (dialog.getFile () == null || dialog.getFile ().length () == 0) {
          new MessageDialog (this, "no file name given for save");
        } else {
          final String file_name = dialog.getDirectory () +
            java.io.File.separator + dialog.getFile ();

          if (new File (file_name).exists ()) {
            final YesNoDialog yes_no_dialog =
              new YesNoDialog (this, "this file exists: " + file_name +
                               " overwrite it?");

            if (yes_no_dialog.getResult ()) {
              // yes - continue
            } else {
              // no
              return;
            }
          }

          System.out.println ("trying to save to: " +  file_name);

          final MessageDialog message;

          if (entry.getFeatureCount () > 200) {
            message = new MessageDialog (this, "saving ...", false);
          } else {
            message = null;
          }

          if (include_diana_extensions) {
            entry.save (file_name);
          } else {
            entry.saveEMBLOnly (file_name);
          }

          if (keep_new_name) {
            entry.setName (file_name);
          }

          if (message != null) {
            message.dispose ();
          }
        }
      } else {
        final MessageDialog message;

        if (entry.getFeatureCount () > 200) {
          message = new MessageDialog (this, "saving...", false);
        } else {
          message = null;
        }

        if (include_diana_extensions) {
          entry.save ();
        } else {
          entry.saveEMBLOnly ();
        }

        if (message != null) {
          message.dispose ();
        }
      }
    } catch (ReadOnlyException e) {
      new MessageDialog (this, "this entry is read only");
    } catch (IOException e) {
      new MessageDialog (this, "error while saving: " + e);
    }

    if (invalid_features != null && invalid_features.size () > 0) {
      final MessageDialog message =
        new MessageDialog (this,
                           "warning: some features may have invalid " +
                           "start or stop codons (these have been selected)",
                           false);

      getSelection ().set (invalid_features);
    }
  }

  /**
   *  Save the changes to all the Entry objects in the entry_group back to
   *  where the they came from.
   **/
  public void saveAllEntries () {
    for (int entry_index = 0 ;
         entry_index < entry_group.size () ;
         ++entry_index) {
      saveEntry (entry_group.elementAt (entry_index), true, false, true);
    }
  }

  /**
   *  Read the features from a file choosen by the user into the default entry.
   *  This method uses EntryFileDialog to open a requester.
   **/
  public void readFeatures () {
    if (getEntryGroup ().getDefaultEntry () == null) {
      new MessageDialog (this, "there is no default entry");
      return;
    }

    final Entry default_entry = getEntryGroup ().getDefaultEntry ();

    final EntryFileDialog dialog = new EntryFileDialog (this);

    // get the entry and add it to entry_group
    final embl.Entry new_embl_entry = dialog.getEntry ();

    if (new_embl_entry == null) {
      return;
    }

    final Entry entry =
      new Entry (getEntryGroup ().getBases (), new_embl_entry);

    final FeatureVector new_features = entry.getAllFeatures ();

    for (int i = 0 ; i < new_features.size () ; ++i) {
      final Feature this_feature = new_features.elementAt (i);

      this_feature.removeFromEntry ();

      default_entry.add (this_feature);
    }
  }

  /**
   *  Call pack () and then move the Frame to the centre of the screen.
   **/
  public void packme () {
    pack ();

    final Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();

    final int x_position = (screen.width - getSize ().width) / 2;
    int y_position = (screen.height - getSize ().height) / 2;

    if (y_position < 10) {
      y_position = 10;
    }

    setLocation (new Point (x_position, y_position));
  }

  /**
   *  Make and add the menus for this component.
   **/
  private void makeMenus () {
    final Font default_font = getDefaultFont ();

    setMenuBar (menu_bar);

//    menu_bar.setFont (default_font);

    if (Diana.isStandAlone ()) {
      // the applet can't save or read

      final MenuItem read_features = new MenuItem ("Read Features");
      read_features.addActionListener (new ActionListener () {
        public void actionPerformed (ActionEvent event) {
          readFeatures ();
        }
      });

      file_menu.add (read_features);

      final MenuItem read_entry = new MenuItem ("Read An Entry");
      read_entry.addActionListener (new ActionListener () {
        public void actionPerformed (ActionEvent event) {
          readEntry ();
        }
      });

      file_menu.add (read_entry);

      file_menu.addSeparator ();

      final MenuItem clone_entry_edit = new MenuItem ("Clone This Window");
      clone_entry_edit.addActionListener (new ActionListener () {
        public void actionPerformed (ActionEvent event) {
          Diana.makeEntryEdit (getEntryGroup ()).show ();
        }
      });

      file_menu.add (clone_entry_edit);

      file_menu.addSeparator ();

      final MenuItem save_default = new MenuItem ("Save Default Entry");
      save_default.addActionListener (new ActionListener () {
        public void actionPerformed (ActionEvent event) {
          saveDefaultEntry ();
        }
      });

      file_menu.add (save_default);

      final MenuItem save = new MenuItem ("Save An Entry ...");
      save.addActionListener (new ActionListener () {
        public void actionPerformed (ActionEvent event) {
          saveAnEntry ();
        }
      });

      file_menu.add (save);

      final MenuItem save_as = new MenuItem ("Save An Entry As ...");
      save_as.addActionListener (new ActionListener () {
        public void actionPerformed (ActionEvent event) {
          saveAnEntryAs ();
        }
      });

      file_menu.add (save_as);

      final MenuItem save_embl = new MenuItem ("Save An Entry As EMBL ...");
      save_embl.addActionListener (new ActionListener () {
        public void actionPerformed (ActionEvent event) {
          saveAnEntryAsEMBL ();
        }
      });

      file_menu.add (save_embl);

      final MenuItem save_all = new MenuItem ("Save All Entries");
      save_all.addActionListener (new ActionListener () {
        public void actionPerformed (ActionEvent event) {
          saveAllEntries ();
        }
      });

      file_menu.add (save_all);

      file_menu.addSeparator ();

      final MenuItem set_default = new MenuItem ("Set Default Entry ...");
      set_default.addActionListener (new ActionListener () {
        public void actionPerformed (ActionEvent event) {
          setDefaultEntry ();
        }
      });

      file_menu.add (set_default);

      final MenuItem set_default_entry_name =
        new MenuItem ("Set Name Of Default Entry");
      set_default_entry_name.addActionListener (new ActionListener () {
        public void actionPerformed (ActionEvent event) {
          setDefaultEntryName ();
        }
      });

      file_menu.add (set_default_entry_name);

      file_menu.addSeparator ();
    }

    final MenuItem close = new MenuItem ("Close");
    close.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        entryEditFinished ();
      }
    });

    file_menu.add (close);

    menu_bar.add (file_menu);


    final CheckboxMenuItem show_base_display_item =
      new CheckboxMenuItem ("Show Base View");
    show_base_display_item.setState (true);
    show_base_display_item.addItemListener (new ItemListener () {
      public void itemStateChanged(ItemEvent event) {
        base_display.setVisible (show_base_display_item.getState ());
        packme ();
      }
    });
    display_menu.add (show_base_display_item);

    final CheckboxMenuItem show_feature_list_item =
      new CheckboxMenuItem ("Show Feature List");
    show_feature_list_item.setState (feature_list.isVisible ());
    show_feature_list_item.addItemListener (new ItemListener () {
      public void itemStateChanged(ItemEvent event) {
        feature_list.setVisible (show_feature_list_item.getState ());
        packme ();
      }
    });
    display_menu.add (show_feature_list_item);


    display_menu.addSeparator ();

    if (Diana.isStandAlone ()) {
      final MenuItem forward_usage_plot_item =
        new MenuItem ("Add A Forward Strand Usage Plot");
      forward_usage_plot_item.addActionListener (new ActionListener () {
        public void actionPerformed (ActionEvent event) {
          addUsagePlot (true);
        }
      });
      display_menu.add (forward_usage_plot_item);

      final MenuItem reverse_usage_plot_item =
        new MenuItem ("Add A Reverse Strand Usage Plot");
      reverse_usage_plot_item.addActionListener (new ActionListener () {
        public void actionPerformed (ActionEvent event) {
          addUsagePlot (false);
          packme ();
        }
      });
      display_menu.add (reverse_usage_plot_item);

      display_menu.addSeparator ();
    }


    final BaseAlgorithm [] algorithms = base_plot_group.getPlotAlgorithms ();

    for (int i = 0 ; i < algorithms.length ; ++i) {
      final BaseAlgorithm this_algorithm = algorithms[i];

      addDisplayMenuAlgorithm (this_algorithm);
    }


    select_menu = new SelectMenu (this, this, getEntryGroup ());
    menu_bar.add (select_menu);

    view_menu = new ViewMenu (this, this, getEntryGroup ());
    menu_bar.add (view_menu);

    goto_menu = new GotoMenu (this, this, this, getEntryGroup ());
    menu_bar.add (goto_menu);

    if (Diana.isStandAlone ()) {
      edit_menu = new EditMenu (this, this, getEntryGroup ());
      menu_bar.add (edit_menu);

      add_menu = new AddMenu (this, this, getEntryGroup ());
      menu_bar.add (add_menu);

      write_menu = new WriteMenu (this, this, getEntryGroup ());
      menu_bar.add (write_menu);

      run_menu = new RunMenu (this, this);
      menu_bar.add (run_menu);
    }

    menu_bar.add (display_menu);
  }

  /**
   *  Ask the user for a file name, then read the codon usage data from that
   *  file, then make and add a BasePlot component using the data.
   **/
  private void addUsagePlot (final boolean use_forward_strand) {
    final FileDialog dialog =
      new FileDialog (this,
                      "Select a codon usage data file name ...",
                      FileDialog.LOAD);

    dialog.setVisible (true);

    if (dialog.getFile () != null && dialog.getFile ().length () != 0) {
      final String codon_usage_file_name =
        dialog.getDirectory () + java.io.File.separator + dialog.getFile ();

      try {
        final BaseAlgorithm new_algorithm;

        if (use_forward_strand) {
          final Strand forward_strand =
            getEntryGroup ().getBases ().getForwardStrand ();

          final CodonUsageWeight usage_weights =
            new CodonUsageWeight (codon_usage_file_name, forward_strand);

          new_algorithm =
            new CodonUsageAlgorithm (forward_strand, usage_weights);
        } else {
          final Strand backward_strand =
            getEntryGroup ().getBases ().getBackwardStrand ();

          final CodonUsageWeight usage_weights =
            new CodonUsageWeight (codon_usage_file_name, backward_strand);

          new_algorithm =
            new CodonUsageAlgorithm (backward_strand, usage_weights);
        }

        base_plot_group.addAlgorithm (new_algorithm);

        base_plot_group.setVisibleByAlgorithm (new_algorithm, true);

        packme ();

        addDisplayMenuAlgorithm (new_algorithm);

        feature_display.fireAdjustmentEvent ();
      } catch (IOException e) {
        new MessageDialog (this, "error while reading: " + e);
      }
    }

  }

  /**
   *  Add a menu item for the given algorithm to the Display menu.
   **/
  private void addDisplayMenuAlgorithm (final BaseAlgorithm algorithm) {
    final CheckboxMenuItem new_item =
      new CheckboxMenuItem (algorithm.getAlgorithmName ());

    new_item.setState (base_plot_group.basePlotIsVisible (algorithm));

    new_item.addItemListener (new ItemListener () {
      public void itemStateChanged(ItemEvent event) {
        base_plot_group.setVisibleByAlgorithm (algorithm,
                                               new_item.getState ());
        packme ();
      }
    });

    display_menu.add (new_item);
  }

  /**
   *  Return the current default font (from Diana.options).
   **/
  private Font getDefaultFont () {
    return Diana.getOptions ().getFont ();
  }

  /**
   *  A vector containing the Entry objects that this EntryEdit object knows
   *  about.  The first Entry object in the entry_group is special - see
   *  saveEntry () for details.
   **/
  private EntryGroup entry_group;

  private final Panel top_panel = new Panel ();
  private final MenuBar menu_bar = new MenuBar ();
  private final Menu file_menu = new Menu ("File");
  private Menu select_menu;
  private Menu view_menu;
  private Menu goto_menu;
  private Menu edit_menu;
  private Menu add_menu;
  private Menu write_menu;
  private Menu run_menu;
  private Menu graph_menu;
  private final Menu display_menu = new Menu ("Display");

  private EntryGroupDisplay group_display;
  private FeatureDisplay feature_display;
  private FeatureDisplay base_display;
  private BasePlotGroup base_plot_group;
  private SelectionInfoDisplay selection_info;

  private FeatureList feature_list;
//  private final FeatureList feature_list2 = new FeatureList (this, 15);

  /**
   *  This Object contains the current selection.
   **/
  private Selection selection = new Selection ();

  /**
   *  A vector of those objects listening for selection change events.
   **/
  final private Vector goto_listener_list = new Vector ();
}


