/* FeaturePlotGroup.java
 *
 * created: Wed Dec 16 1998
 *
 * This file is part of Artemis
 * 
 * Copyright (C) 1998,1999  Genome Research Limited
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/diana/components/FeaturePlotGroup.java,v 1.11 1999/03/22 11:00:24 kmr Exp $
 */

package diana.components;

import Diana;
import diana.*;
import diana.sequence.*;
import diana.plot.*;
import java.awt.*;
import java.awt.event.*;

/**
 *  This is a super-component containing several FeaturePlot components, each
 *  of which can toggled off and on.  The component contains a row of toggle
 *  buttons and then the FeaturePlot components below.
 *
 *  @author Kim Rutherford
 *  @version $Id: FeaturePlotGroup.java,v 1.11 1999/03/22 11:00:24 kmr Exp $
 **/

public class FeaturePlotGroup extends Frame
    implements EntryChangeListener {
  /**
   *  Create a new FeaturePlotGroup component for the given feature.
   **/
  public FeaturePlotGroup (Feature feature) {
    super ("Graphs for: " + feature.getIDString ());
    this.feature = feature;
    this.entry = feature.getEntry ();

    // don't repeat an algorithm in this array.
    final FeatureAlgorithm [] plot_value_producers = {
      new HydrophobicityAlgorithm (getFeature ()),
      new HydrophilicityAlgorithm (getFeature ()),
      new CoilFeatureAlgorithm (getFeature ())
    };

    final Font font = Diana.getOptions ().getFont ();

    setFont (font);

    GridBagLayout gridbag = new GridBagLayout();
    setLayout (gridbag);

    GridBagConstraints c = new GridBagConstraints();

    c.fill = GridBagConstraints.HORIZONTAL;
    c.anchor = GridBagConstraints.NORTH;
    c.gridwidth = GridBagConstraints.REMAINDER;
    c.gridheight = 1;
    c.weightx = 1;
    c.weighty = 0;

    final FlowLayout flow_layout = new FlowLayout (FlowLayout.LEFT); 
    final Panel button_panel = new Panel (flow_layout);
    gridbag.setConstraints (button_panel, c);
    add (button_panel);

    final Label new_label = new Label ("Feature Algorithms: ");
    button_panel.add (new_label);

    c.fill = GridBagConstraints.BOTH;
    c.weighty = 1;
    c.insets = new Insets (0,0,5,0);
    
    for (int i = 0 ; i < plot_value_producers.length ; ++i) {
      final FeatureAlgorithm this_algorithm = plot_value_producers[i];

      final boolean show_algorithm = true;
      
      // create a button and a FeaturePlot component for each algorithm.  the
      // FeaturePlot is initially hidden.
      final Checkbox new_checkbox =
        new Checkbox (this_algorithm.getAlgorithmName (), show_algorithm);

      new_checkbox.addItemListener (new ItemListener () {
        public void itemStateChanged (ItemEvent event) {
          final String check_label =
            ((Checkbox) event.getSource ()).getLabel ();

          setVisibleByName (check_label,
                            event.getStateChange () == ItemEvent.SELECTED);
        }
      });

      button_panel.add (new_checkbox);

      final FeaturePlot new_feature_plot = new FeaturePlot (this_algorithm);

      gridbag.setConstraints (new_feature_plot, c);
      add (new_feature_plot);
      new_feature_plot.setVisible (show_algorithm);
    }

    getFeature ().getEntry ().addEntryChangeListener (this);

    addWindowListener (new WindowAdapter () {
      public void windowClosing (WindowEvent event) {
        stopListening ();
        FeaturePlotGroup.this.dispose ();
      }
    });

    bottom_button_panel = new Panel ();
    c.fill = GridBagConstraints.HORIZONTAL;
    c.weighty = 0;
    gridbag.setConstraints (bottom_button_panel, c);
    add (bottom_button_panel);

    close_button = new Button ("Close");
    close_button.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent e) {
        stopListening ();
        FeaturePlotGroup.this.dispose ();
      }
    });

    bottom_button_panel.add (close_button);
    
    pack ();

    int new_x_size = feature.getTranslation ().length () * 2;

    if (new_x_size > 1000) {
      new_x_size = 1000;
    }

    final int new_y_size =
      200 * plot_value_producers.length + button_panel.getSize ().height;
    
    // give each FeaturePlot component a height of 200
    setSize (new_x_size, new_y_size);

    final Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
    setLocation (new Point ((screen.width - getSize ().width) / 2,
                            (screen.height - getSize ().height) / 2));

    show ();
  }

  /**
   *  Remove this object as a entry change listener and then call
   *  stopListening () on each FeaturePlot component.
   **/
  private void stopListening () {
    getEntry ().removeEntryChangeListener (this);

    final Component [] children = getComponents ();

    for (int i = 0 ; i < children.length ; ++i) {
      if (children[i] instanceof FeaturePlot) {
        ((FeaturePlot)children[i]).stopListening ();
      }
    }
  }
  
  /**
   *  Implementation of the EntryChangeListener interface.  We listen to
   *  EntryChange events so we can delete this component if the feature gets
   *  deleted.
   **/
  public void entryChanged (EntryChangeEvent event) {
//     System.out.println ("in FeaturePlotGroup: entry change event type: " +
//                         event.getType () + " for entry " +
//                         event.getSource () + " for feature " +
//                         event.getFeature ());
    switch (event.getType ()) {
    case EntryChangeEvent.FEATURE_DELETED:
      if (event.getFeature () == getFeature ()) {
        stopListening ();
        dispose ();
      }
      break;
    default:
      // do nothing;
      break;
    }
  }

  /**
   *  Find the Checkbox component in this FeaturePlotGroup object that has the
   *  given label.
   **/
  private Component findCheckbox (String find_label) {
    final Component [] children = getComponents ();

    for (int i = 0 ; i < children.length ; ++i) {
      if (children[i] instanceof FeaturePlot) {
        final String component_algorithm_name =
          ((FeaturePlot)children[i]).getAlgorithm ().getAlgorithmName ();
        if (component_algorithm_name.equals (find_label)) {
          return children[i];
        }
      }
    }

    return null;
  }

  /**
   *  Given an algorithm name, find and set the visibility of the
   *  corresponding FeaturePlot component.
   **/
  private void setVisibleByName (String label, boolean visible) {
    final Component plot = FeaturePlotGroup.this.findCheckbox (label);

    plot.setSize (plot.getSize ().width, 100);
    plot.setVisible (visible);
    pack ();
  }

  /**
   *  Return the feature that this component is plotting.
   **/
  private Feature getFeature () {
    return feature;
  }

  /**
   *  Return the Entry that contains the Feature this object is displaying.
   **/
  private Entry getEntry () {
    return entry;
  }

  /**
   *  The feature we are plotting.
   **/
  private Feature feature;

  /**
   *  The Entry that contains the Feature this object is displaying.
   **/
  private Entry entry;

  /**
   *  Pressing this button will distroy the Frame.
   **/
  private Button close_button;
  
  /**
   *  A Panel to hold the buttons.
   **/
  private Panel bottom_button_panel;
}
