/* GeoCodonWindowAlgorithm.java
 *
 * created: Thu Mar 25 1999
 *
 * This file is part of Artemis
 * 
 * Copyright (C) 1999  Genome Research Limited
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/diana/plot/GeoCodonWindowAlgorithm.java,v 1.2 1999/05/04 10:00:37 kmr Exp $
 */

package diana.plot;

import OutOfRangeException;
import diana.sequence.*;
import embl.Range;

/**
 *  Objects of this class have one useful method - getValues (), which takes a
 *  range of bases and returns three floating point numbers.  The return
 *  values is calculated by comparing the bases to the weight matrix that is
 *  passed to constructor.  The return values from getValues () are the
 *  geometric averages of the weights of the codons in the range.
 *
 *  @author Kim Rutherford
 *  @version $Id: GeoCodonWindowAlgorithm.java,v 1.2 1999/05/04 10:00:37 kmr Exp $
 **/

public class GeoCodonWindowAlgorithm extends BaseAlgorithm {
  /**
   *  Create a new GeoCodonWindowAlgorithm object.  See getValues () for
   *  detail on how the weightings array is used.
   *  @param strand The Strand to do the calculation on.
   *  @param name The name of the algorithm.
   *  @param weightings The weight matrix to use in getValues ().  The weights
   *    should appear in this order: ttt, ttc, tta, ttg, tct, ..., ggg.
   **/
  public GeoCodonWindowAlgorithm (final Strand strand,
                                  final String name,
                                  final float [] weightings) {
    super (strand, name);

    this.weightings = weightings;
  }

  /**
   *  Return the geometric average of the user weightings for the codons in
   *  the given window.
   *  @param start The start base (included in the range).
   *  @param end The end base (included in the range).
   *  @param values The three return values for this algorithm are returned in
   *    this array.  There is one value for each frame and the value is the
   *    the geometric average of the weightings for the codons in the range.
   **/
  public void getValues (int start, int end, float [] values) {
    // add 1 or 2 if necessary to make the range a multiple of 3
    if (getStrand ().isForwardStrand ()) {
      end -= (end - start + 1) % 3;
    } else {
      start += (end - start + 1) % 3;
    }

    final String sequence;

    try {
      sequence = getStrand ().getRawSubSequence (new Range (start, end));
    } catch (OutOfRangeException e) {
      throw new Error ("internal error - unexpected exception: " + e);
    }

    float [] totals = { 1, 1, 1 };

    // a count of the number of codons we have seen
    int [] codon_count = { 0, 0, 0 };

    final int sub_sequence_length = sequence.length ();

    if (getStrand ().isForwardStrand ()) {

      for (int frame = 0 ; frame < 3 ; ++frame) {
        for (int i = frame ; i < sub_sequence_length - 3 ; i += 3) {

          final char base1 = sequence.charAt (i);
          final char base2 = sequence.charAt (i + 1);
          final char base3 = sequence.charAt (i + 2);

          final int index =
            Bases.getIndexOfBase (base1) * 16 +
            Bases.getIndexOfBase (base2) * 4 +
            Bases.getIndexOfBase (base3);

          final float this_weight = weightings [index];

          if (this_weight > 0) {
            ++codon_count[frame];
            totals[frame] *= this_weight;
          }
        }
      }
    } else {
    }


   System.out.println ("start: " + start + " end: " + end + " returning: " +
                        sequence.length () + " " + totals[0]); 

    for (int frame = 0 ; frame < 3 ; ++frame) {
      if (codon_count[frame] == 0) {
        values[frame] = 0;
      } else {
        values[frame] =
          (float) Math.pow (totals[frame], 1.0 / codon_count[frame]);
      }
    }
  }
  
  /**
   *  Return the number of values a call to getValues () will return - one
   *  in this case.
   **/
  public int getValueCount () {
    return 3;
  }

  /**
   *  The weightings array that was passed to the constructor.
   **/
  float [] weightings = null;
}

