/* Qualifier.java
 *
 * created: Tue Oct 13 1998
 *
 * This file is part of Artemis
 *
 * Copyright (C) 1998,1999  Genome Research Limited
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/embl/Qualifier.java,v 1.18 1999/04/27 14:51:05 kmr Exp $
 */

package embl;

import StringVector;

/**
 *  Each object of this class represents a group of embl qualifiers
 *  (name/value pairs) with the same name, so in effect the object contains
 *  one name and many values.
 *
 *  @author Kim Rutherford
 *  @version $Id: Qualifier.java,v 1.18 1999/04/27 14:51:05 kmr Exp $
 * */

public class Qualifier {
  /**
   *  Create a new Qualifier object.  This object consists of a name and
   *  possibly some values.  In the raw embl file we have /name=value.
   *  @param name The name of this qualifier (ie. the text immediately after
   *    the / in the qualifier)
   * @param values The values of this qualifier (ie the text immediately after
   *    the = in the qualifiers).  This argument should be null if the
   *    qualifier has no values.  The values String objects should not include
   *    the quote characters (),"" or [].  For example if the original
   *    qualifier was /citation=[3] then the value String should be: 3.  This
   *    argument is copied by the constructor.
   *  @exception QualifierParseException Thrown if the format of any of the
   *    value String is not appropriate for a Qualifier with the given name.
   *    Each qualifier has a specific format for the value part which depends
   *    on the name, for example the value part of /codon_start qualifier must
   *    be a number: 1, 2 or 3.
   *  @exception InvalidQualifierException Thrown if the name to the
   *    constructor is not a valid feature qualifier name.
   *  @exception InvalidQualifierException Thrown if the name String is not a
   *    valid embl qualifier name, if the value argument is null when there
   *    should be a value or if it has a value when it should be null
   **/
  public Qualifier (final String name, final StringVector values)
      throws QualifierParseException, InvalidQualifierException {
    initialise (name, values);
  }

  /**
   *  Create a new Qualifier object.  This object consists of a name and
   *  possibly some values.  In the raw embl file we have /name=value.
   *  @param name The name of this qualifier (ie. the text immediately after
   *    the / in the qualifier)
   *  @param value The value of this qualifier (see the other constructor for
   *    details).  Other values may be added later.
   *  @exception QualifierParseException Thrown if the format of the
   *    value String is not appropriate for a Qualifier with the given name.
   *    Each qualifier has a specific format for the value part which depends
   *    on the name, for example the value part of /codon_start qualifier must
   *    be a number: 1, 2 or 3.  Also thrown if a qualifier has value when it
   *    should not or vice versa.
   *  @exception InvalidQualifierException Thrown if the name String is not a
   *    valid embl qualifier name, if the value argument is null when there
   *    should be a value or if it has a value when it should be null
   **/
  public Qualifier (final String name, final String value)
      throws QualifierParseException, InvalidQualifierException {
    initialise (name, new StringVector (value));
  }

  /**
   *  Create a new Qualifier object.  This object will consist of just a name.
   *  In the raw embl file we have /name.
   *  @param name The name of this qualifier (ie. the text immediately after
   *    the / in the qualifier)
   *  @exception QualifierParseException Thrown if the format of the
   *    value String is not appropriate for a Qualifier with the given name.
   *    Each qualifier has a specific format for the value part which depends
   *    on the name, for example the value part of /codon_start qualifier must
   *    be a number: 1, 2 or 3.  Also thrown if a qualifier has value when it
   *    should not or vice versa.
   *  @exception InvalidQualifierException Thrown if the name String is not a
   *    valid embl qualifier name, if the value argument is null when there
   *    should be a value or if it has a value when it should be null
   **/
  public Qualifier (final String name)
      throws QualifierParseException, InvalidQualifierException {
    initialise (name, (StringVector) null);
  }

  /**
   *  Called by the constructor to initialise the object.
   *  @param name The name of this qualifier (ie. the text immediately after
   *    the / in the qualifier)
   *  @param values The values of this qualifier (see the other constructor for
   *    details).  Other values may be added later.
   *  @exception QualifierParseException Thrown if the format of any of the
   *    value String is not appropriate for a Qualifier with the given name.
   *    Each qualifier has a specific format for the value part which depends
   *    on the name, for example the value part of /codon_start qualifier must
   *    be a number: 1, 2 or 3.  Also thrown if a qualifier has value when it
   *    should not or vice versa.
   *  @exception InvalidQualifierException Thrown if the name String is not a
   *    valid embl qualifier name, if the value argument is null when there
   *    should be a value or if it has a value when it should be null
   **/
  private void initialise (final String name, final StringVector values)
      throws QualifierParseException, InvalidQualifierException {

    this.name = EntryInformation.checkQualifier (name);

    if (this.name == null) {
      throw new InvalidQualifierException ("\"" + name + "\" cannot be used " +
                                           "as a qualifier");
    }

    if (values == null) {
      this.values = null;
    } else {
      if (values.size () == 0) {
        throw new Error ("internal error - zero length values vector");
      }

      this.values = values.copy ();
    }

    final QualifierInfo qualifier_info =
      EntryInformation.getQualifierInfo (name);

    if (qualifier_info.getType () == QualifierInfo.OPTIONAL) {
      // no need to check the value
    } else {
      if (qualifier_info.getType () == QualifierInfo.NO_VALUE &&
          !allValuesNull ()) {
        throw new InvalidQualifierException ("a /" + name + " qualifier " +
                                             "must not have a value");
      }

      if (qualifier_info.getType () == QualifierInfo.QUOTED_TEXT &&
          !allValuesNotNull ()) {
        // a QUOTED_TEXT qualifier must have a value even if the value is only
        // an empty string
        throw new InvalidQualifierException ("a /" + name + " qualifier " +
                                             "must have a value");
      }

      if (qualifier_info.getType () != QualifierInfo.NO_VALUE &&
          qualifier_info.getType () != QualifierInfo.QUOTED_TEXT &&
          (values == null || values.elementAt (0).equals (""))) {
        throw new InvalidQualifierException ("a /" + name + " qualifier " +
                                             "must have a value");
      }
    }

    checkFormat ();
  }

  /**
   *  Return true if and only if all the values of this Qualifier are null.
   **/
  private boolean allValuesNull () {
    if (values == null) {
      return true;
    } else {
      for (int i = 0 ; i < values.size () ; ++i) {
        if (values.elementAt (i) != null) {
          return false;
        }
      }

      return true;
    }
  }

  /**
   *  Return true if and only if all the values of this Qualifier are not null.
   **/
  private boolean allValuesNotNull () {
    if (values == null) {
      return false;
    } else {
      for (int i = 0 ; i < values.size () ; ++i) {
        if (values.elementAt (i) == null) {
          return false;
        }
      }

      return true;
    }
  }
  /**
   *  Return the name of this qualifier.
   **/
  public String getName () {
    return name;
  }

  /**
   *  Return the values of this qualifier.  The StringVector that is returned
   *  is a copy (use addValues (), removeValue () etc. to change the
   *  Qualifier), but the String objects inside the vector are not copied so
   *  the same String references will occur in all copies.
   **/
  public StringVector getValues () {
    if (values == null) {
      return null;
    } else {
      return values.copy ();
    }
  }

  /**
   *  Add the given values to this object.
   **/
  public void addValues (final StringVector new_values) {
    if (values == null) {
      values = new StringVector ();
      values.add ((String)null);
    }
    if (new_values == null) {
      // if new_values is null then we have a qualifier with no value (like
      // /pseudo).  we add one null value for each occurrence of the
      // qualifier.
      values.add ((String)null);
    } else {
      for (int i = 0 ; i < new_values.size () ; ++i) {
        values.add (new_values.elementAt (i));
      }
    }
  }

  /**
   *  Add the given value to this object.
   **/
  public void addValue (final String new_value) {
    if (values == null) {
      values = new StringVector ();
    }
//       System.out.println ("adding " + new_value);
    values.add (new_value);
  }

  /**
   *  Remove the given value from this object.  The reference of the argument
   *  String must have previously been passed to addValue () or addValues ()
   *  for a removal to occur.
   **/
  public void removeValue (final String value) {
//       System.out.println ("adding " + new_values.elementAt (i));
    values.remove (value);

    if (values.size () == 0) {
      values = null;
    }
  }

  /**
   *  Return the reference of a new copy of this Qualifier.
   **/
  public Qualifier copy () {
    try {
      return new Qualifier (getName (), getValues ());
    } catch (QualifierParseException e) {
      throw new Error ("internal error - " +
                       "illegal state in Qualifier.copy (): " + e);
    } catch (InvalidQualifierException e) {
      throw new Error ("internal error - " +
                       "illegal state in Qualifier.copy (): " + e);
    }
  }

  /**
   *  Return a StringVector containing the possible qualifiers for the given
   *  key.
   *  @see embl.EntryInformation
   **/
  public static StringVector getPossibleQualifiers (final Key key) {
    return EntryInformation.getPossibleQualifierNames (key);
  }

  /**
   *  Return true if and only if given qualifier_name is a legal qualifier
   *  name for the given key.
   *  @see embl.EntryInformation
   **/
  public static boolean isPossibleQualifier (final Key key,
                                             final String qualifier_name) {
    return EntryInformation.isPossibleQualifier (key, qualifier_name);
  }

  /**
   *  Return a String reference containing the qualifier_name if
   *  qualifier_name is a valid qualifier name, otherwise returns null.  The
   *  returned reference refers to a single object, so, for example, passing
   *  any String containing "gene" will always return the same reference (to a
   *  String object containing "gene", whose reference is stored in this
   *  package).
   **/
  public static String checkQualifier (final String qualifier_name) {
    return EntryInformation.checkQualifier (qualifier_name);
  }

  /**
   *  Return true if and only if the format of the values StringVector of this
   *  Qualifier is appropriate for a Qualifier with this name.  Each qualifier
   *  has a specific format for the value part which depends on the name, for
   *  example the value part of /codon_start qualifier must be a number: 1, 2
   *  or 3.
   *  @exception QualifierParseException Thrown if the format of any of the
   *    value String is not appropriate for a Qualifier with the given name.
   *    Each qualifier has a specific format for the value part which depends
   *    on the name, for example the value part of /codon_start qualifier must
   *    be a number: 1, 2 or 3.  Also thrown if a qualifier has value when it
   *    should not or vice versa.
   **/
  private void checkFormat ()
      throws QualifierParseException {
    if (false) {
      throw new QualifierParseException ("incorrect qualifier format for: " +
                                         this);
    }
  }

  /**
   *  The name that was passed to the constructor.
   **/
  private String name;

  /**
   *  The values that were passed to the constructor.
   **/
  private StringVector values;
}


